/*
 *    vmg - a versatile multigrid solver
 *    Copyright (C) 2012 Institute for Numerical Simulation, University of Bonn
 *
 *  vmg is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  vmg is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * interface_gaussian.cpp
 *
 *  Created on: 19.04.2013
 *      Author: Julian Iseringhausen
 */

#ifdef HAVE_CONFIG_H
#include <libvmg_config.h>
#endif

#include <boost/math/special_functions/erf.hpp>

#include <cassert>
#include <cmath>
#include <iostream>
#include <limits>

#include "base/math.hpp"
#include "grid/grid.hpp"
#include "grid/multigrid.hpp"

#include "interface_gaussian.hpp"

using namespace VMG;
using VMGInterfaces::InterfaceGaussian;

void InterfaceGaussian::ImportRightHandSide(Multigrid& multigrid)
{
  Index i;
  Vector pos;

  Grid& grid = multigrid(multigrid.MaxLevel());
  grid.Clear();

  const Vector pos_center = 0.5 * (grid.Extent().Begin() + grid.Extent().End());

  const Index begin = grid.Local().Begin() + grid.Local().Size() / 4;
  const Index end = grid.Local().End() - grid.Local().Size() / 4;

  for (i.X()=begin.X(); i.X()<end.X(); ++i.X())
    for (i.Y()=begin.Y(); i.Y()<end.Y(); ++i.Y())
      for (i.Z()=begin.Z(); i.Z()<end.Z(); ++i.Z()) {
        const vmg_float dist_sqr = std::pow((grid.GetSpatialPos(i) - pos_center).Length(), 2);
	grid(i) = std::exp(-1.0 * dist_sqr / (2.0 * sigma * sigma)) / std::pow(sigma * std::sqrt(2 * VMG::Math::pi), 3);
      }
}

void InterfaceGaussian::ExportSolution(Grid& grid)
{
  Index i;
  vmg_float err_1 = 0.0;
  vmg_float err_2 = 0.0;
  vmg_float err_inf = 0.0;

  const Vector pos_center = 0.5 * (grid.Extent().Begin() + grid.Extent().End());

  for (i.X()=grid.Local().Begin().X(); i.X()<grid.Local().End().X(); ++i.X())
    for (i.Y()=grid.Local().Begin().Y(); i.Y()<grid.Local().End().Y(); ++i.Y())
      for (i.Z()=grid.Local().Begin().Z(); i.Z()<grid.Local().End().Z(); ++i.Z()) {

        const vmg_float dist = (grid.GetSpatialPos(i) - pos_center).Length();

	if (dist > std::numeric_limits<vmg_float>::epsilon()) {

	  const vmg_float err = std::abs(grid.GetVal(i) - 1.0 / (4.0 * VMG::Math::pi * dist) *
                                boost::math::erf(dist / (VMG::Math::sqrt2 * sigma)));

	  err_1 += err;
	  err_2 += err * err;
	  err_inf = std::max(err_inf, err);

	}
      }

  err_1 = grid.Extent().MeshWidth().Product() * err_1;
  err_2 = std::sqrt(grid.Extent().MeshWidth().Product() * err_2);
  err_inf = grid.Extent().MeshWidth().Product() * err_inf;

  std::cout << std::scientific << "Error L1-Norm:   " << err_1 << std::endl
	    << "Error L2-Norm:   " << err_2 << std::endl
	    << "Error Inf-Norm: " << err_inf << std::endl;
}
